"""
    saltrewrite.testsuite.fix_slow_test_decorator
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Replaces any use of ``@slowTest`` with ``@pytest.mark.slow_test``,
    and, in case ``pytest`` isn't yet imported, it additionall adds the missing
    import
"""
from bowler import Query
from saltrewrite import utils

MARKER = "pytest.mark.slow_test"
DECORATOR = "slowTest"


def rewrite(paths, interactive=False, silent=False):
    """
    Rewrite the passed in paths
    """
    # Don't waste time on non-test files
    paths = utils.filter_test_files(paths)
    if not paths:
        return
    (
        Query(paths)
        .select("classdef|funcdef")
        .filter(filter_not_decorated)
        .modify(replace_decorator)
        .execute(write=True, interactive=interactive, silent=silent)
    )


def filter_not_decorated(node, capture, filename):
    """
    Filter undecorated nodes
    """
    return bool(utils.get_decorator(node, DECORATOR, MARKER))


def replace_decorator(node, capture, filename):
    """
    Replaces usage of ``@slowTest`` with ``@pytest.mark.slow_test``
    """
    return utils.rewrite_decorator(node, DECORATOR, MARKER)



"""
    saltrewrite.fixes
    ~~~~~~~~~~~~~~~~~

    Entry point for all the salt rewrite fixes
"""
import operator
from collections import OrderedDict

import saltrewrite.imports
import saltrewrite.salt
import saltrewrite.testsuite


class RegistryClass:
    """
    Registry class to hold all available fixes
    """

    __slots__ = ("__fixes__",)

    def __init__(self):
        __fixes__ = (
            saltrewrite.imports.__fixes__
            + saltrewrite.testsuite.__fixes__
            + saltrewrite.salt.__fixes__
        )
        __sorted_fixes__ = (
            (module.__name__.split(".")[-1], module)
            for (priority, module) in sorted(__fixes__, key=operator.itemgetter(0))
        )
        self.__fixes__ = OrderedDict(__sorted_fixes__)

    def fixes(self, excluded_names=(), only_names=()):
        """
        Returns all available fixes, optionally skipping those passed in `excluded_names`
        """
        for name in self.__fixes__:
            if only_names:
                if name in only_names:
                    yield name, self.__fixes__[name]
                continue
            if name in excluded_names:
                continue
            yield name, self.__fixes__[name]

    def fix_names(self):
        """
        Returns the list of the fix names
        """
        return list(self.__fixes__)


Registry = RegistryClass()  # pylint: disable=invalid-name



"""
    saltrewrite.utils
    ~~~~~~~~~~~~~~~~~

    @todo: add description
"""
import os

from bowler import SYMBOL
from bowler import TOKEN
from bowler.helpers import find_first
from fissix import fixer_util
from fissix import pygram
from fissix.fixer_util import parenthesize
from fissix.fixer_util import touch_import
from fissix.pygram import python_symbols as syms
from fissix.pytree import Leaf


def get_indent(node):
    """
    Returns the indent for the passed node
    """
    indent = None
    while node:
        indent = find_first(node, TOKEN.INDENT)
        if indent is not None:
            indent = indent.value
            break
        node = node.parent

    return indent


def remove_import(node, name):
    """
    Removes the import.
    """

    # def is_import_stmt(node):
    #    return (
    #        node.type == fixer_util.syms.simple_stmt  # pylint: disable=no-member
    #        and node.children
    #        and fixer_util.is_import(node.children[0])
    #    )

    root = fixer_util.find_root(node)
    if "." in name:
        raise NotImplementedError("Dotted imports not yet supported")

    import_node = fixer_util.find_binding(name, root)
    if import_node:
        import_node.remove()


def get_from_imports(node):
    """
    Return `from module import bar` imports
    """
    from_imports = []
    for child in node.children:
        if child.type != pygram.python_symbols.import_as_names:  # pylint: disable=no-member
            continue
        for leaf in child.children:
            if leaf.type == TOKEN.NAME:
                from_imports.append(leaf)
    return from_imports


def remove_from_import(node, package, name):
    """
    Removes the ``from X import Y`` from the module.
    """

    root = fixer_util.find_root(node)

    import_node = fixer_util.find_binding(name, root, package)
    if not import_node:
        return
    from_imports = get_from_imports(import_node)

    from_imports_children = []
    while from_imports:
        from_import = from_imports.pop(0)
        if from_import.value == name:
            from_import.remove()
            continue
        from_imports_children.append(from_import)

    if not from_imports_children:
        import_node.remove()
        return

    new_import = fixer_util.FromImport(package, from_imports_children)
    import_node.replace(new_import)


def is_multiline(node):
    """
    Checks if node is a multiline string
    """
    if isinstance(node, list):
        return any(is_multiline(n) for n in node)

    for leaf in node.leaves():
        if "\n" in leaf.prefix:
            return True
    return False


def parenthesize_if_necessary(node):
    """
    Parenthesize node if necessary to avoid syntax errors
    """
    if is_multiline(node):
        # If not already parenthesized, parenthesize
        for first_leaf in node.leaves():
            if first_leaf.type in (TOKEN.LPAR, TOKEN.LBRACE, TOKEN.LSQB):
                # Already parenthesized
                return node
            break
        return parenthesize(node.clone())
    return node


def keyword(name, **kwargs):
    """
    A helper to produce keyword nodes
    """
    kwargs.setdefault("prefix", " ")
    return Leaf(TOKEN.NAME, name, **kwargs)


def filter_test_files(paths):
    """
    Filter paths which don't match a salt test module
    """
    if not isinstance(paths, (list, tuple)):
        paths = [paths]
    _paths = []
    for path in paths:
        if not os.path.basename(path).startswith("test_"):
            continue
        _paths.append(path)
    return _paths


def get_decorator(node, decorator_name, marker):
    """
    Don't modify classes or test methods that aren't decorated with ``DECORATOR``
    """
    if node.parent.type == SYMBOL.decorated:
        child = node.parent.children[0]
        if child.type == SYMBOL.decorator:
            decorators = [child]
        elif child.type == SYMBOL.decorators:
            decorators = child.children
        else:
            raise NotImplementedError
        for decorator in decorators:
            name = decorator.children[1]
            assert name.type in {TOKEN.NAME, SYMBOL.dotted_name}

            if str(name) in (decorator_name, marker):
                return decorator


def rewrite_decorator(node, decorator_name, marker):
    """
    Replaces usage of ``@<decorator_name>`` with ``@<marker>``
    """
    decorator = get_decorator(node, decorator_name, marker)

    # pylint: disable=no-member

    for leaf in decorator.children:
        if leaf.type == TOKEN.AT:
            continue
        if leaf.type == TOKEN.NAME and leaf.value == decorator_name:
            leaf.value = marker
            continue
        if leaf.type == syms.atom:
            # This scenario is func(*args)
            #
            # Sometimes instead of passing expanded arguments 'foo(arg1, arg2)' a tuple or a
            # list is passed as first argument 'foo([arg1, arg2])' or 'foo((arg1, arg2)).
            # Let's break out of those containers.
            break_out_of_atom = False
            for atom_child in leaf.children:
                if atom_child.type not in (syms.listmaker, syms.dictsetmaker, syms.testlist_gexp):
                    atom_child.remove()
                    continue
                break_out_of_atom = True
                break
            if break_out_of_atom:
                continue
        elif leaf.type == syms.arglist:
            # This scenario is func(*args, **kwargs)
            #
            for arg_child in leaf.children:
                if arg_child.type == syms.atom:
                    # Sometimes instead of passing expanded arguments 'foo(arg1, arg2)' a tuple or a
                    # list is passed as first argument 'foo([arg1, arg2])' or 'foo((arg1, arg2)).
                    # Let's break out of those containers.
                    break_out_of_atom = False
                    for atom_child in arg_child.children:
                        if atom_child.type not in (
                            syms.listmaker,
                            syms.dictsetmaker,
                            syms.testlist_gexp,
                        ):
                            atom_child.remove()
                            continue
                        break_out_of_atom = True
                        break
                    if break_out_of_atom:
                        continue

    # pylint: enable=no-member

    touch_import(None, "pytest", node)
    remove_from_import(node, "tests.support.helpers", decorator_name)



"""
    saltrewrite.salt.fix_dunder_utils
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Fix ``__utils__["*.*"](...)`` calls to import and call the real utils module function.
"""
import ast
import os
import pathlib
from functools import lru_cache

import click
from bowler import Query
from bowler import SYMBOL
from bowler import TOKEN
from bowler.types import Leaf
from bowler.types import Node
from fissix.fixer_util import Call
from fissix.fixer_util import Dot
from fissix.fixer_util import touch_import


SALT_DUNDERS = (
    "__active_provider_name__",
    "__context__",
    "__env__",
    "__events__",
    "__executors__",
    "__grains__",
    "__instance_id__",
    "__jid_event__",
    "__low__",
    "__lowstate__",
    "__master_opts__",
    "__opts__",
    "__pillar__",
    "__proxy__",
    "__reg__",
    "__ret__",
    "__runner__",
    "__running__",
    "__salt__",
    "__salt_system_encoding__",
    "__serializers__",
    "__states__",
    "__utils__",
)


class DunderParser(ast.NodeTransformer):  # pylint: disable=missing-class-docstring
    # pylint: disable=missing-function-docstring,invalid-name
    def __init__(self):
        self.virtualname = None
        self.uses_salt_dunders = False

    def visit_Name(self, node):
        if node.id in SALT_DUNDERS:
            self.uses_salt_dunders = True
        return self.generic_visit(node)

    def visit_Assign(self, node):
        for target in node.targets:
            if not isinstance(target, ast.Name):
                continue
            if target.id == "__virtualname__":
                self.virtualname = node.value.s
        return self.generic_visit(node)

    # pylint: enable=missing-function-docstring,invalid-name


def _get_salt_code_root():
    return pathlib.Path.cwd()


@lru_cache
def get_utils_module_info():
    """
    Collect utils modules dunder information.
    """
    mapping = {}
    salt_utils_package_path = _get_salt_code_root().joinpath("salt", "utils")
    for path in salt_utils_package_path.rglob("*.py"):
        transformer = DunderParser()
        tree = ast.parse(path.read_text())
        transformer.visit(tree)
        mapping[path.resolve()] = {
            "modname": path.stem,
            "virtualname": transformer.virtualname or path.stem,
            "uses_salt_dunders": transformer.uses_salt_dunders,
        }
    return mapping


@lru_cache
def get_utils_module_details(name):
    """
    Return utils module details.
    """
    full_module_name = f"salt.utils.{name}"
    full_module_path = pathlib.Path(f"{full_module_name.replace('.', os.sep)}.py").resolve()
    utils_module_info = get_utils_module_info()
    if full_module_path.exists():
        return utils_module_info[full_module_path]
    modname = name.split(".")[0]
    for entry in utils_module_info.values():
        if entry["virtualname"] == modname:
            return entry
    raise RuntimeError(
        f"Could not find the python module for {name!r} and '{full_module_path}' does not exist"
    )


def rewrite(paths, interactive=False, silent=False):
    """
    Rewrite the passed in paths
    """
    (
        Query(paths)
        .select(
            """
            (
                dunder_call=power<
                    '__utils__'
                    trailer< '[' dunder_mod_func=any* ']' >
                    trailer< '(' function_arguments=any* ')' >
                    trailing=any*
                >
            )
            """
        )
        .modify(fix_dunder_utils_calls)
        .execute(write=True, interactive=interactive, silent=silent)
    )


def fix_dunder_utils_calls(node, capture, filename):
    """
    Automatically rewrite dunder utils calls to call the module directly.
    """
    if "dunder_mod_func" not in capture:
        return
    dunder_mod_func = capture["dunder_mod_func"][0].value.strip("'").strip('"')

    utils_module, utils_module_funcname = dunder_mod_func.split(".")
    details = get_utils_module_details(utils_module)
    if details["uses_salt_dunders"]:

        click.echo(
            f" * Not calling 'salt.utils.{details['modname']}.{utils_module_funcname}' "
            f"directly because internally the 'salt.utils.{details['modname']}' "
            "module uses salt dunders",
            err=True,
        )
        return

    # Make sure we import the right utils module
    touch_import(None, f"salt.utils.{details['modname']}", node)
    click.echo(f" * Fixing dunder module func call: '{dunder_mod_func}'")

    # Un-parent the function arguments so we can add them to a new call
    for leaf in capture["function_arguments"]:
        leaf.parent = None

    # Un-parent any trailign code after the __utils__ call too
    for leaf in capture["trailing"]:
        leaf.parent = None

    # Create the new function call
    call_node = Call(
        Leaf(TOKEN.NAME, utils_module_funcname, prefix=""),
        capture["function_arguments"],
    )

    # Create replacement node
    replacement = Node(
        SYMBOL.power,
        [
            Leaf(TOKEN.NAME, "salt", prefix=capture["node"].prefix),
            Node(
                SYMBOL.trailer,
                [
                    Dot(),
                    Leaf(TOKEN.NAME, "utils", prefix=""),
                    Dot(),
                    Leaf(TOKEN.NAME, details["modname"], prefix=""),
                    Dot(),
                    call_node,
                ],
            ),
            *capture["trailing"],
        ],
    )
    # Replace the whole node with the new function call
    node.replace(replacement)


# pylint: enable=missing-function-docstring



# pylint: disable=missing-module-docstring
from saltrewrite.imports import fix_tornado_imports

__all__ = [modname for modname in list(globals()) if modname.startswith("fix_")]

__fixes__ = [
    (getattr(module, "FIX_PRIORITY", 0), module)
    for module in [globals()[modname] for modname in __all__]
]



# pylint: disable=missing-module-docstring
from saltrewrite.testsuite import fix_asserts
from saltrewrite.testsuite import fix_destructive_test_decorator
from saltrewrite.testsuite import fix_expensive_test_decorator
from saltrewrite.testsuite import fix_requires_network_decorator
from saltrewrite.testsuite import fix_requires_salt_modules_decorator
from saltrewrite.testsuite import fix_requires_salt_states_decorator
from saltrewrite.testsuite import fix_skip_if_binaries_missing_decorator
from saltrewrite.testsuite import fix_skip_if_not_root_decorator
from saltrewrite.testsuite import fix_slow_test_decorator

__all__ = [modname for modname in list(globals()) if modname.startswith("fix_")]

__fixes__ = [
    (getattr(module, "FIX_PRIORITY", 0), module)
    for module in [globals()[modname] for modname in __all__]
]



"""
    saltrewrite.salt.fix_docstrings
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    @todo: add description
"""
# pylint: disable=consider-using-f-string
import operator
import re
import sys
from collections import namedtuple
from functools import total_ordering

from bowler import Query
from bowler import TOKEN


def rewrite(paths, interactive=False, silent=False):
    """
    Rewrite the passed in paths
    """
    (
        Query(paths)
        .select(
            """
            (
                file_input<
                    simple_stmt<
                        [STRING]
                        docstring=any*
                    >
                    any*
                >
            |
                class_def=classdef<
                    any*
                    suite<
                        any*
                        simple_stmt<
                            [STRING]
                            docstring=any*
                        >
                        any*
                    >
                    any*
                >
            |
                funcdef<
                    any*
                    suite<
                        any*
                        simple_stmt<
                            [STRING]
                            docstring=any*
                        >
                        any*
                    >
                    any*
                >
            )
            """
        )
        .filter(filter_no_module_docstrings)
        .modify(fix_module_docstrings)
        .execute(write=True, interactive=interactive, silent=silent)
    )


def filter_no_module_docstrings(node, capture, filename):
    """
    If the first child is a docstring, process it
    """
    if "docstring" not in capture:
        # Docstring was not captured, ignore
        return
    if capture["docstring"][0].type != TOKEN.STRING:
        # It's not a docstring, return
        return
    # Process node
    return node


def fix_module_docstrings(node, capture, filename):
    """
    Automaticaly run fixes against docstrings
    """
    # Replace the docstring Leaf value with a fixed docstring
    capture["docstring"][0].value = autofix_docstring(capture["docstring"][0].value, filename)


def autofix_docstring(docstring, filename):
    """
    Run auto fix code against the docstring
    """
    return _convert_version_names_to_numbers(
        _fix_codeblocks(
            _fix_directives_formatting(
                _fix_simple_cli_example_spacing_issues(docstring, filename), filename
            ),
            filename,
        ),
        filename,
    )


CONVERT_VERSION_NAMES_TO_NUMBERS_RE = re.compile(
    r"\.\.(?:[ ]*)?((?P<vtype>(versionadded|versionchanged|deprecated))(?:[:]+)(?:[ ]+)?(?P<version>[^\n]*))"
)


def _handle_convert_version_names_to_numbers_match(match):
    """
    Convert every match with a properly formatted version.
    """
    vtype = match.group("vtype")
    version = match.group("version")
    versions = set()
    splitters = (",", "/", " ")
    for splitter in splitters:
        for _vs in version.split(splitter):
            for _splitter in splitters:
                if _splitter in _vs:
                    break
            else:
                versions.add(_vs.strip())
    parsed_versions = []
    for version in versions:
        try:
            parsed_versions.append(SaltStackVersion.from_name(version))
        except ValueError:
            try:
                parsed_versions.append(SaltStackVersion.parse(version))
            except ValueError as exc:
                raise RuntimeError(f"Unable to parse {version!r} as a SaltStackVersion") from exc

    replace_contents = ".. {}:: {}".format(
        vtype, ",".join([v.string for v in sorted(parsed_versions)])
    )
    return replace_contents


def _convert_version_names_to_numbers(docstring, filename):
    """
    Convert Salt version names to their version numbers counterpart
    """
    return CONVERT_VERSION_NAMES_TO_NUMBERS_RE.sub(
        _handle_convert_version_names_to_numbers_match, docstring
    )


CLI_EXAMPLE_CASE_AND_SPACING_RE = re.compile(
    r"(?:[\n\r]+)([ ]+)CLI Example(?P<plural>s)?(?:[\s]+)?:(?:[^\n\r]+)?(?:[\n\r]+)",
    flags=re.I | re.MULTILINE,
)
CLI_EXAMPLE_MISSING_CODE_BLOCK_RE = re.compile(
    r"\n([ ]+)CLI Example(?P<plural>s)?:\n\n([\s]+)salt ", flags=re.I | re.MULTILINE
)


def _fix_simple_cli_example_spacing_issues(docstring, filename):
    """
    Fix spacing for expected Salt CLI examples
    """
    return CLI_EXAMPLE_MISSING_CODE_BLOCK_RE.sub(
        r"\n\1CLI Example\2:\n\n\1..code-block:: bash\n\n\3salt ",
        CLI_EXAMPLE_CASE_AND_SPACING_RE.sub(r"\n\n\1CLI Example\2:\n\n", docstring),
    )


DIRECTIVES_FORMATTING_RE = re.compile(
    r"(\n(?P<spc1>[ ]+)?((?P<dots>[.]{2,})(?P<spc2>[ ]+)?"
    r"(?P<directive>(?:[^ :]+)))(?:[:]{2})(?P<spc3>[ ]+)?"
    r"(?P<remaining>[^\n]+)?\n)"
)


def _handle_fix_directives_formatting_match(match):
    return (
        "\n{}.. {}:: {}".format(
            match.group("spc1") or "",
            match.group("directive"),
            match.group("remaining") or "",
        ).rstrip()
        + "\n"
    )


def _fix_directives_formatting(docstring, filename):
    """
    Fix directive definition spacing
    """
    return DIRECTIVES_FORMATTING_RE.sub(_handle_fix_directives_formatting_match, docstring)


FIX_CODE_BLOCKS_RE = re.compile(
    r"^(?P<spc1>[ ]+)?(?P<dots>[.]{2}) (?P<directive>code-block)::(?P<lang>.*)\n$"
)


def _fix_codeblocks(docstring, filename):
    """
    Fix expected empty lines around code blocks
    """
    output = []
    found_codeblock = False
    for line in docstring.splitlines(True):
        match = FIX_CODE_BLOCKS_RE.match(line)
        if found_codeblock:
            if line.strip() and line.strip().startswith(":"):
                # code block directive configuration
                output.append(line)
                continue
            if line.strip():
                # We need an empty line after the code-block
                output.append("\n")
            found_codeblock = False
        if match:
            found_codeblock = True
        output.append(line)
    return "".join(output)


# pylint: disable=missing-function-docstring,missing-class-docstring

MAX_SIZE = sys.maxsize
VERSION_LIMIT = MAX_SIZE - 200


@total_ordering
class SaltVersion(namedtuple("SaltVersion", "name, info, released")):  # pragma: no cover
    __slots__ = ()

    def __new__(cls, name, info, released=False):
        if isinstance(info, int):
            info = (info,)
        return super().__new__(cls, name, info, released)

    def __eq__(self, other):
        return self.info == other.info

    def __gt__(self, other):
        return self.info > other.info


class SaltVersionsInfo(type):  # pragma: no cover

    _sorted_versions = ()
    _current_release = None
    _previous_release = None
    _next_release = None

    # ----- Please refrain from fixing whitespace ---------------------------------->
    # The idea is to keep this readable.
    # -------------------------------------------------------------------------------
    # fmt: off
    HYDROGEN      = SaltVersion("Hydrogen"     , info=(2014, 1),  released=True)
    HELIUM        = SaltVersion("Helium"       , info=(2014, 7),  released=True)
    LITHIUM       = SaltVersion("Lithium"      , info=(2015, 5),  released=True)
    BERYLLIUM     = SaltVersion("Beryllium"    , info=(2015, 8),  released=True)
    BORON         = SaltVersion("Boron"        , info=(2016, 3),  released=True)
    CARBON        = SaltVersion("Carbon"       , info=(2016, 11), released=True)
    NITROGEN      = SaltVersion("Nitrogen"     , info=(2017, 7),  released=True)
    OXYGEN        = SaltVersion("Oxygen"       , info=(2018, 3),  released=True)
    FLUORINE      = SaltVersion("Fluorine"     , info=(2019, 2),  released=True)
    NEON          = SaltVersion("Neon"         , info=3000,       released=True)
    SODIUM        = SaltVersion("Sodium"       , info=3001,       released=True)
    MAGNESIUM     = SaltVersion("Magnesium"    , info=3002,       released=True)
    ALUMINIUM     = SaltVersion("Aluminium"    , info=3003,       released=True)
    SILICON       = SaltVersion("Silicon"      , info=3004,       released=True)
    PHOSPHORUS    = SaltVersion("Phosphorus"   , info=3005,       released=True)
    SULFUR        = SaltVersion("Sulfur"       , info=(3006, 0),  released=True)
    CHLORINE      = SaltVersion("Chlorine"     , info=(3007, 0))
    ARGON         = SaltVersion("Argon"        , info=(3008, 0))
    POTASSIUM     = SaltVersion("Potassium"    , info=(3009, 0))
    CALCIUM       = SaltVersion("Calcium"      , info=(3010, 0))
    SCANDIUM      = SaltVersion("Scandium"     , info=(3011, 0))
    TITANIUM      = SaltVersion("Titanium"     , info=(3012, 0))
    VANADIUM      = SaltVersion("Vanadium"     , info=(3013, 0))
    CHROMIUM      = SaltVersion("Chromium"     , info=(3014, 0))
    MANGANESE     = SaltVersion("Manganese"    , info=(3015, 0))
    IRON          = SaltVersion("Iron"         , info=(3016, 0))
    COBALT        = SaltVersion("Cobalt"       , info=(3017, 0))
    NICKEL        = SaltVersion("Nickel"       , info=(3018, 0))
    COPPER        = SaltVersion("Copper"       , info=(3019, 0))
    ZINC          = SaltVersion("Zinc"         , info=(3020, 0))
    GALLIUM       = SaltVersion("Gallium"      , info=(3021, 0))
    GERMANIUM     = SaltVersion("Germanium"    , info=(3022, 0))
    ARSENIC       = SaltVersion("Arsenic"      , info=(3023, 0))
    SELENIUM      = SaltVersion("Selenium"     , info=(3024, 0))
    BROMINE       = SaltVersion("Bromine"      , info=(3025, 0))
    KRYPTON       = SaltVersion("Krypton"      , info=(3026, 0))
    RUBIDIUM      = SaltVersion("Rubidium"     , info=(3027, 0))
    STRONTIUM     = SaltVersion("Strontium"    , info=(3028, 0))
    YTTRIUM       = SaltVersion("Yttrium"      , info=(3029, 0))
    ZIRCONIUM     = SaltVersion("Zirconium"    , info=(3030, 0))
    NIOBIUM       = SaltVersion("Niobium"      , info=(3031, 0))
    MOLYBDENUM    = SaltVersion("Molybdenum"   , info=(3032, 0))
    TECHNETIUM    = SaltVersion("Technetium"   , info=(3033, 0))
    RUTHENIUM     = SaltVersion("Ruthenium"    , info=(3034, 0))
    RHODIUM       = SaltVersion("Rhodium"      , info=(3035, 0))
    PALLADIUM     = SaltVersion("Palladium"    , info=(3036, 0))
    SILVER        = SaltVersion("Silver"       , info=(3037, 0))
    CADMIUM       = SaltVersion("Cadmium"      , info=(3038, 0))
    INDIUM        = SaltVersion("Indium"       , info=(3039, 0))
    TIN           = SaltVersion("Tin"          , info=(3040, 0))
    ANTIMONY      = SaltVersion("Antimony"     , info=(3041, 0))
    TELLURIUM     = SaltVersion("Tellurium"    , info=(3042, 0))
    IODINE        = SaltVersion("Iodine"       , info=(3043, 0))
    XENON         = SaltVersion("Xenon"        , info=(3044, 0))
    CESIUM        = SaltVersion("Cesium"       , info=(3045, 0))
    BARIUM        = SaltVersion("Barium"       , info=(3046, 0))
    LANTHANUM     = SaltVersion("Lanthanum"    , info=(3047, 0))
    CERIUM        = SaltVersion("Cerium"       , info=(3048, 0))
    PRASEODYMIUM  = SaltVersion("Praseodymium" , info=(3049, 0))
    NEODYMIUM     = SaltVersion("Neodymium"    , info=(3050, 0))
    PROMETHIUM    = SaltVersion("Promethium"   , info=(3051, 0))
    SAMARIUM      = SaltVersion("Samarium"     , info=(3052, 0))
    EUROPIUM      = SaltVersion("Europium"     , info=(3053, 0))
    GADOLINIUM    = SaltVersion("Gadolinium"   , info=(3054, 0))
    TERBIUM       = SaltVersion("Terbium"      , info=(3055, 0))
    DYSPROSIUM    = SaltVersion("Dysprosium"   , info=(3056, 0))
    HOLMIUM       = SaltVersion("Holmium"      , info=(3057, 0))
    ERBIUM        = SaltVersion("Erbium"       , info=(3058, 0))
    THULIUM       = SaltVersion("Thulium"      , info=(3059, 0))
    YTTERBIUM     = SaltVersion("Ytterbium"    , info=(3060, 0))
    LUTETIUM      = SaltVersion("Lutetium"     , info=(3061, 0))
    HAFNIUM       = SaltVersion("Hafnium"      , info=(3062, 0))
    TANTALUM      = SaltVersion("Tantalum"     , info=(3063, 0))
    TUNGSTEN      = SaltVersion("Tungsten"     , info=(3064, 0))
    RHENIUM       = SaltVersion("Rhenium"      , info=(3065, 0))
    OSMIUM        = SaltVersion("Osmium"       , info=(3066, 0))
    IRIDIUM       = SaltVersion("Iridium"      , info=(3067, 0))
    PLATINUM      = SaltVersion("Platinum"     , info=(3068, 0))
    GOLD          = SaltVersion("Gold"         , info=(3069, 0))
    MERCURY       = SaltVersion("Mercury"      , info=(3070, 0))
    THALLIUM      = SaltVersion("Thallium"     , info=(3071, 0))
    LEAD          = SaltVersion("Lead"         , info=(3072, 0))
    BISMUTH       = SaltVersion("Bismuth"      , info=(3073, 0))
    POLONIUM      = SaltVersion("Polonium"     , info=(3074, 0))
    ASTATINE      = SaltVersion("Astatine"     , info=(3075, 0))
    RADON         = SaltVersion("Radon"        , info=(3076, 0))
    FRANCIUM      = SaltVersion("Francium"     , info=(3077, 0))
    RADIUM        = SaltVersion("Radium"       , info=(3078, 0))
    ACTINIUM      = SaltVersion("Actinium"     , info=(3079, 0))
    THORIUM       = SaltVersion("Thorium"      , info=(3080, 0))
    PROTACTINIUM  = SaltVersion("Protactinium" , info=(3081, 0))
    URANIUM       = SaltVersion("Uranium"      , info=(3082, 0))
    NEPTUNIUM     = SaltVersion("Neptunium"    , info=(3083, 0))
    PLUTONIUM     = SaltVersion("Plutonium"    , info=(3084, 0))
    AMERICIUM     = SaltVersion("Americium"    , info=(3085, 0))
    CURIUM        = SaltVersion("Curium"       , info=(3086, 0))
    BERKELIUM     = SaltVersion("Berkelium"    , info=(3087, 0))
    CALIFORNIUM   = SaltVersion("Californium"  , info=(3088, 0))
    EINSTEINIUM   = SaltVersion("Einsteinium"  , info=(3089, 0))
    FERMIUM       = SaltVersion("Fermium"      , info=(3090, 0))
    MENDELEVIUM   = SaltVersion("Mendelevium"  , info=(3091, 0))
    NOBELIUM      = SaltVersion("Nobelium"     , info=(3092, 0))
    LAWRENCIUM    = SaltVersion("Lawrencium"   , info=(3093, 0))
    RUTHERFORDIUM = SaltVersion("Rutherfordium", info=(3094, 0))
    DUBNIUM       = SaltVersion("Dubnium"      , info=(3095, 0))
    SEABORGIUM    = SaltVersion("Seaborgium"   , info=(3096, 0))
    BOHRIUM       = SaltVersion("Bohrium"      , info=(3097, 0))
    HASSIUM       = SaltVersion("Hassium"      , info=(3098, 0))
    MEITNERIUM    = SaltVersion("Meitnerium"   , info=(3099, 0))
    DARMSTADTIUM  = SaltVersion("Darmstadtium" , info=(3100, 0))
    ROENTGENIUM   = SaltVersion("Roentgenium"  , info=(3101, 0))
    COPERNICIUM   = SaltVersion("Copernicium"  , info=(3102, 0))
    NIHONIUM      = SaltVersion("Nihonium"     , info=(3103, 0))
    FLEROVIUM     = SaltVersion("Flerovium"    , info=(3104, 0))
    MOSCOVIUM     = SaltVersion("Moscovium"    , info=(3105, 0))
    LIVERMORIUM   = SaltVersion("Livermorium"  , info=(3106, 0))
    TENNESSINE    = SaltVersion("Tennessine"   , info=(3107, 0))
    OGANESSON     = SaltVersion("Oganesson"    , info=(3108, 0))
    # <---- Please refrain from fixing whitespace -----------------------------------
    # The idea is to keep this readable.
    # -------------------------------------------------------------------------------
    # fmt: on

    @classmethod
    def versions(cls):
        if not cls._sorted_versions:
            cls._sorted_versions = sorted(
                (getattr(cls, name) for name in dir(cls) if name.isupper()),
                key=operator.attrgetter("info"),
            )
        return cls._sorted_versions

    @classmethod
    def current_release(cls):
        if cls._current_release is None:
            for version in cls.versions():
                if version.released is False:
                    cls._current_release = version
                    break
        return cls._current_release

    @classmethod
    def next_release(cls):
        if cls._next_release is None:
            next_release_ahead = False
            for version in cls.versions():
                if next_release_ahead:
                    cls._next_release = version
                    break
                if version == cls.current_release():
                    next_release_ahead = True
        return cls._next_release

    @classmethod
    def previous_release(cls):
        if cls._previous_release is None:
            previous = None
            for version in cls.versions():
                if version == cls.current_release():
                    break
                previous = version
            cls._previous_release = previous
        return cls._previous_release


class SaltStackVersion:  # pragma: no cover
    """
    Handle SaltStack versions class.

    Knows how to parse ``git describe`` output, knows about release candidates
    and also supports version comparison.
    """

    __slots__ = (
        "name",
        "major",
        "minor",
        "bugfix",
        "mbugfix",
        "pre_type",
        "pre_num",
        "noc",
        "sha",
    )

    git_sha_regex = r"(?P<sha>g?[a-f0-9]{7,40})"

    git_describe_regex = re.compile(
        r"(?:[^\d]+)?(?P<major>[\d]{1,4})"
        r"(?:\.(?P<minor>[\d]{1,2}))?"
        r"(?:\.(?P<bugfix>[\d]{0,2}))?"
        r"(?:\.(?P<mbugfix>[\d]{0,2}))?"
        r"(?:(?P<pre_type>rc|a|b|alpha|beta|nb)(?P<pre_num>[\d]+))?"
        r"(?:(?:.*)(?:\+|-)(?P<noc>(?:0na|[\d]+|n/a))(?:-|\.)" + git_sha_regex + r")?"
    )
    git_sha_regex = r"^" + git_sha_regex

    git_sha_regex = re.compile(git_sha_regex)

    NAMES = {v.name: v.info for v in SaltVersionsInfo.versions()}
    LNAMES = {k.lower(): v for (k, v) in iter(NAMES.items())}
    VNAMES = {v: k for (k, v) in iter(NAMES.items())}
    RMATCH = {v[:2]: k for (k, v) in iter(NAMES.items())}

    def __init__(
        self,  # pylint: disable=C0103
        major,
        minor=None,
        bugfix=None,
        mbugfix=0,
        pre_type=None,
        pre_num=None,
        noc=0,
        sha=None,
    ):
        if isinstance(major, str):
            major = int(major)

        if isinstance(minor, str):
            if not minor:
                # Empty string
                minor = None
            else:
                minor = int(minor)
        if self.can_have_dot_zero(major):
            minor = minor if minor else 0

        if bugfix is None and not self.new_version(major=major):
            bugfix = 0
        elif isinstance(bugfix, str):
            if not bugfix:
                bugfix = None
            else:
                bugfix = int(bugfix)

        if mbugfix is None:
            mbugfix = 0
        elif isinstance(mbugfix, str):
            mbugfix = int(mbugfix)

        if pre_type is None:
            pre_type = ""
        if pre_num is None:
            pre_num = 0
        elif isinstance(pre_num, str):
            pre_num = int(pre_num)

        if noc is None:
            noc = 0
        elif isinstance(noc, str) and noc in ("0na", "n/a"):
            noc = -1
        elif isinstance(noc, str):
            noc = int(noc)

        self.major = major
        self.minor = minor
        self.bugfix = bugfix
        self.mbugfix = mbugfix
        self.pre_type = pre_type
        self.pre_num = pre_num
        if self.can_have_dot_zero(major):
            vnames_key = (major, 0)
        elif self.new_version(major):
            vnames_key = (major,)
        else:
            vnames_key = (major, minor)
        self.name = self.VNAMES.get(vnames_key)
        self.noc = noc
        self.sha = sha

    def new_version(self, major):
        """
        determine if using new versioning scheme
        """
        return bool(int(major) >= 3000 and int(major) < VERSION_LIMIT)

    def can_have_dot_zero(self, major):
        """
        determine if using new versioning scheme
        """
        if int(major) < 3000:
            return True
        return bool(int(major) >= 3006 and int(major) < VERSION_LIMIT)

    @classmethod
    def parse(cls, version_string):
        if version_string.lower() in cls.LNAMES:
            return cls.from_name(version_string)
        vstr = version_string.decode() if isinstance(version_string, bytes) else version_string
        match = cls.git_describe_regex.match(vstr)
        if not match:
            raise ValueError(f"Unable to parse version string: '{version_string}'")
        return cls(*match.groups())

    @classmethod
    def from_name(cls, name):
        if name.lower() not in cls.LNAMES:
            raise ValueError(f"Named version '{name}' is not known")
        return cls(*cls.LNAMES[name.lower()])

    @classmethod
    def current_release(cls):
        return cls(*SaltVersionsInfo.current_release().info)

    @classmethod
    def next_release(cls):
        return cls(*SaltVersionsInfo.next_release().info)

    @property
    def sse(self):
        # Higher than 0.17, lower than first date based
        return 0 < self.major < 2014

    def min_info(self):
        info = [self.major]
        if self.new_version(self.major):
            if self.minor:
                info.append(self.minor)
            elif self.can_have_dot_zero(self.major):
                info.append(self.minor)
        else:
            info.extend([self.minor, self.bugfix, self.mbugfix])
        return info

    @property
    def info(self):
        return tuple(self.min_info())

    @property
    def pre_info(self):
        info = self.min_info()
        info.extend([self.pre_type, self.pre_num])
        return tuple(info)

    @property
    def noc_info(self):
        info = self.min_info()
        info.extend([self.pre_type, self.pre_num, self.noc])
        return tuple(info)

    @property
    def full_info(self):
        info = self.min_info()
        info.extend([self.pre_type, self.pre_num, self.noc, self.sha])
        return tuple(info)

    @property
    def full_info_all_versions(self):
        """
        Return the full info regardless
        of which versioning scheme we
        are using.
        """
        info = [
            self.major,
            self.minor,
            self.bugfix,
            self.mbugfix,
            self.pre_type,
            self.pre_num,
            self.noc,
            self.sha,
        ]
        return tuple(info)

    @property
    def string(self):
        if self.new_version(self.major):
            version_string = f"{self.major}"
            if self.minor:
                version_string = f"{self.major}.{self.minor}"
            if not self.minor and self.can_have_dot_zero(self.major):
                version_string = f"{self.major}.{self.minor}"
        else:
            version_string = f"{self.major}.{self.minor}.{self.bugfix}"
        if self.mbugfix:
            version_string += f".{self.mbugfix}"
        if self.pre_type:
            version_string += f"{self.pre_type}{self.pre_num}"
        if self.noc and self.sha:
            noc = self.noc
            if noc < 0:
                noc = "0na"
            version_string += f"+{noc}.{self.sha}"
        return version_string

    @property
    def formatted_version(self):
        if self.name and self.major > 10000:
            version_string = self.name
            if self.sse:
                version_string += " Enterprise"
            version_string += " (Unreleased)"
            return version_string
        version_string = self.string
        if self.sse:
            version_string += " Enterprise"
        if (self.major, self.minor) in self.RMATCH:
            version_string += f" ({self.RMATCH[(self.major, self.minor)]})"
        return version_string

    @property
    def pre_index(self):
        if self.new_version(self.major):
            pre_type = 2
            if not isinstance(self.minor, int):
                pre_type = 1
        else:
            pre_type = 4
        return pre_type

    def __str__(self):
        return self.string

    def __compare__(self, other, method):
        if not isinstance(other, SaltStackVersion):
            if isinstance(other, str):
                other = SaltStackVersion.parse(other)
            elif isinstance(other, (list, tuple)):
                other = SaltStackVersion(*other)
            else:
                raise ValueError(f"Cannot instantiate Version from type '{type(other)}'")
        pre_type = self.pre_index
        other_pre_type = other.pre_index
        other_noc_info = list(other.noc_info)
        noc_info = list(self.noc_info)

        if self.new_version(self.major):
            if self.minor and not other.minor:
                # We have minor information, the other side does not
                if self.minor > 0:
                    other_noc_info[1] = 0

            if not self.minor and other.minor:
                # The other side has minor information, we don't
                if other.minor > 0:
                    noc_info[1] = 0

        if self.pre_type and not other.pre_type:
            # We have pre-release information, the other side doesn't
            other_noc_info[other_pre_type] = "zzzzz"

        if not self.pre_type and other.pre_type:
            # The other side has pre-release information, we don't
            noc_info[pre_type] = "zzzzz"

        return method(tuple(noc_info), tuple(other_noc_info))

    def __lt__(self, other):
        return self.__compare__(other, lambda _self, _other: _self < _other)

    def __le__(self, other):
        return self.__compare__(other, lambda _self, _other: _self <= _other)

    def __eq__(self, other):
        return self.__compare__(other, lambda _self, _other: _self == _other)

    def __ne__(self, other):
        return self.__compare__(other, lambda _self, _other: _self != _other)

    def __ge__(self, other):
        return self.__compare__(other, lambda _self, _other: _self >= _other)

    def __gt__(self, other):
        return self.__compare__(other, lambda _self, _other: _self > _other)

    def __repr__(self):
        parts = []
        if self.name:
            parts.append(f"name='{self.name}'")
        parts.extend([f"major={self.major}", f"minor={self.minor}"])

        if self.new_version(self.major):
            if not self.can_have_dot_zero(self.major) and not self.minor:
                parts.remove("".join([x for x in parts if re.search("^minor*", x)]))
        else:
            parts.extend([f"bugfix={self.bugfix}"])

        if self.mbugfix:
            parts.append(f"minor-bugfix={self.mbugfix}")
        if self.pre_type:
            parts.append(f"{self.pre_type}={self.pre_num}")
        noc = self.noc
        if noc == -1:
            noc = "0na"
        if noc and self.sha:
            parts.extend([f"noc={noc}", f"sha={self.sha}"])
        return "<{} {}>".format(self.__class__.__name__, " ".join(parts))


# pylint: enable=missing-function-docstring,missing-class-docstring



"""
    saltrewrite.cli
    ~~~~~~~~~~~~~~~

    ``salt-rewrite``'s CLI interface
"""
import click
from saltrewrite.fixes import Registry

try:
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version


@click.command()
@click.argument(
    "paths",
    nargs=-1,
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
    ),
)
@click.option("--silent/--no-silent", "-s/-S", is_flag=True, default=False)
@click.option("--interactive/--no-interactive", "-i/-I", is_flag=True, default=False)
@click.option("--list-fixes", "-l", is_flag=True)
@click.option(
    "-fix",
    "-F",
    type=click.Choice(Registry.fix_names(), case_sensitive=False),
    multiple=True,
)
@click.option(
    "--exclude-fix",
    "-E",
    type=click.Choice(Registry.fix_names(), case_sensitive=False),
    multiple=True,
)
@click.version_option(version=version("salt-rewrite"))
def rewrite(paths, interactive, silent, list_fixes, fix, exclude_fix):
    """
    Main CLI entry-point
    """
    if list_fixes:
        click.echo("Fixes:\n{}".format("\n".join(f" - {fix}" for fix in Registry.fix_names())))
        return

    if fix and exclude_fix:
        raise click.UsageError("The --fix and --exclude-fix are mutually exclusive options")

    for fixname, module in Registry.fixes(excluded_names=exclude_fix, only_names=fix):
        click.echo(f"Running {fixname} ...", err=True)
        module.rewrite(paths, interactive=interactive, silent=silent)



"""
    saltrewrite.imports.fix_tornado_imports
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Fix tornado imports
"""
# pylint: disable=no-member
from bowler import Query
from bowler import SYMBOL
from bowler import TOKEN
from fissix.fixer_util import find_indentation
from fissix.fixer_util import Name
from fissix.pytree import Leaf
from fissix.pytree import Node


def rewrite(paths, interactive=False, silent=False):
    """
    Rewrite the passed in paths
    """
    (
        Query(paths)
        .select_module("tornado")
        .filter(filter_tornado_imports)
        .rename("salt.ext.tornado")
        .select_root()
        .select("classdef|funcdef")
        .filter(filter_not_decorated)
        .modify(replace_decorators)
        .execute(write=True, interactive=interactive, silent=silent)
    )


def filter_tornado_imports(node, capture, filename):
    """
    Filter tornado imports
    """
    for leaf in capture["node"].leaves():
        if leaf.value == "tornado":
            return True


def _get_decorator(node):
    """
    Don't modify classes or test methods that aren't decorated with ``@tornado``
    """
    if node.parent.type == SYMBOL.decorated:
        child = node.parent.children[0]
        if child.type == SYMBOL.decorator:
            decorators = [child]
        elif child.type == SYMBOL.decorators:
            decorators = child.children
        else:
            raise NotImplementedError
        for decorator in decorators:
            name = decorator.children[1]
            assert name.type in {TOKEN.NAME, SYMBOL.dotted_name}

            if str(name).startswith("tornado."):
                return decorator


def filter_not_decorated(node, capture, filename):
    """
    Filter undecorated nodes
    """
    return bool(_get_decorator(node))


def get_decorator_name(decorator):
    """
    Returns the name of the decorator
    """
    name = decorator.children[1]
    assert name.type in {TOKEN.NAME, SYMBOL.dotted_name}
    return str(name)


def replace_decorators(node, capture, filename):
    """
    Replaces usage of ``@tornado.<etc>`` with ``@salt.ext.tornado.<etc>``
    """
    indent = find_indentation(node)

    decorator = _get_decorator(node)
    decorator.remove()

    decorated = Node(
        SYMBOL.decorated,
        [
            Node(
                SYMBOL.decorator,
                [
                    Leaf(TOKEN.AT, "@"),
                    Name(f"salt.ext.{get_decorator_name(decorator)}"),
                    Leaf(TOKEN.NEWLINE, "\n"),
                ],
            )
        ],
        prefix=decorator.prefix,
    )
    node.replace(decorated)
    decorated.append_child(node)

    if indent is not None:
        node.prefix = indent
    else:
        node.prefix = ""



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['saltrewrite',
 'saltrewrite.imports',
 'saltrewrite.salt',
 'saltrewrite.testsuite']

package_data = \
{'': ['*']}

install_requires = \
['bowler>=0.9.0', 'fissix>=21.6,<22.0']

entry_points = \
{'console_scripts': ['salt-rewrite = saltrewrite.__main__:rewrite']}

setup_kwargs = {
    'name': 'salt-rewrite',
    'version': '2.4.4',
    'description': 'A set of Bowler code to rewrite parts of Salt',
    'long_description': None,
    'author': 'Pedro Algarvio',
    'author_email': 'pedro@algarvio.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



"""
    saltrewrite.testsuite.fix_asserts
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This rewrites unittest `self.assert*` into plain assertions

    The source for most of this code is:

        https://github.com/craigds/decrapify/blob/master/pytestify.py
"""
# pylint: disable=no-member,missing-function-docstring
from functools import wraps

from bowler import Query
from bowler import TOKEN
from bowler.types import Leaf
from bowler.types import Node
from fissix.fixer_util import ArgList
from fissix.fixer_util import Attr
from fissix.fixer_util import Call
from fissix.fixer_util import Comma
from fissix.fixer_util import Dot
from fissix.fixer_util import KeywordArg
from fissix.fixer_util import touch_import
from fissix.pygram import python_symbols as syms
from saltrewrite.utils import filter_test_files
from saltrewrite.utils import keyword

# NOTE: these don't take inversions into account.
# Hence why assertNotEqual is a synonym of assertEqual
SYNONYMS = {
    "assertEquals": "assertEqual",
    "failUnlessEqual": "assertEqual",
    "assertNotEqual": "assertEqual",
    "failIfEqual": "assertEqual",
    "assertIsNot": "assertIs",
    "assertNotIn": "assertIn",
    "failUnless": "assertTrue",
    "assert_": "assertTrue",
    "assertFalse": "assertTrue",
    "failIf": "assertTrue",
    "assertIsNotNone": "assertIsNone",
    "assertMultiLineEqual": "assertEqual",
    "assertSequenceEqual": "assertEqual",
    "assertListEqual": "assertEqual",
    "assertTupleEqual": "assertEqual",
    "assertSetEqual": "assertEqual",
    "assertDictEqual": "assertEqual",
    "assertNotIsInstance": "assertIsInstance",
    "assertNotAlmostEqual": "assertAlmostEqual",
    "assertNotRegex": "assertRegex",
}


ARGUMENTS = {
    "assertEqual": 2,
    "assertIs": 2,
    "assertIn": 2,
    "assertGreater": 2,
    "assertLess": 2,
    "assertGreaterEqual": 2,
    "assertLessEqual": 2,
    "assertIsInstance": 2,
    "assertAlmostEqual": 2,
    "assertTrue": 1,
    "assertIsNone": 1,
    "assertRegex": 2,
}


OPERATORS = {
    "assertEqual": Leaf(TOKEN.EQEQUAL, "==", prefix=" "),
    "assertIs": Leaf(TOKEN.NAME, "is", prefix=" "),
    "assertIn": Leaf(TOKEN.NAME, "in", prefix=" "),
    "assertTrue": [],
    "assertIsNone": [
        Leaf(TOKEN.NAME, "is", prefix=" "),
        Leaf(TOKEN.NAME, "None", prefix=" "),
    ],
    "assertGreater": Leaf(TOKEN.GREATER, ">", prefix=" "),
    "assertLess": Leaf(TOKEN.LESS, "<", prefix=" "),
    "assertGreaterEqual": Leaf(TOKEN.GREATEREQUAL, ">=", prefix=" "),
    "assertLessEqual": Leaf(TOKEN.LESSEQUAL, "<=", prefix=" "),
}


# Functions where we invert the operator, or add a 'not'
INVERT_FUNCTIONS = {
    "assertNotEqual",
    "failIfEqual",
    "assertIsNot",
    "assertNotIn",
    "assertFalse",
    "failIf",
    "assertIsNotNone",
    "assertNotIsInstance",
    "assertNotAlmostEqual",
    "assertNotRegex",
}
BOOLEAN_VALUES = ("True", "False")


def rewrite(paths, interactive=False, silent=False):
    """
    Rewrite the passed in paths
    """
    # Don't waste time on non-test files
    paths = filter_test_files(paths)
    if not paths:
        return
    (
        Query(paths)
        .select_class("TestCase")
        # NOTE: You can append as many .select().modify() bits as you want to one query.
        # Each .modify() acts only on the .select[_*]() immediately prior.
        .select_method("assertEqual")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertEquals")
        .modify(callback=assertmethod_to_assert)
        .select_method("failUnlessEqual")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertNotEqual")
        .modify(callback=assertmethod_to_assert)
        .select_method("failIfEqual")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertIs")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertIsNot")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertIn")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertNotIn")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertTrue")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertFalse")
        .modify(callback=assertmethod_to_assert)
        .select_method("assert_")
        .modify(callback=assertmethod_to_assert)
        .select_method("failUnless")
        .modify(callback=assertmethod_to_assert)
        .select_method("failIf")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertIsNone")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertIsNotNone")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertGreater")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertGreaterEqual")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertLess")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertLessEqual")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertIsInstance")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertNotIsInstance")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertDictEqual")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertSetEqual")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertListEqual")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertTupleEqual")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertSequenceEqual")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertMultiLineEqual")
        .modify(callback=assertmethod_to_assert)
        .select_method("assertAlmostEqual")
        .modify(callback=assertalmostequal_to_assert)
        .select_method("assertNotAlmostEqual")
        .modify(callback=assertalmostequal_to_assert)
        .select(
            """
            function_call=power<
                self_attr="self" raises_attr=trailer< "." "assertRaises" >
                trailer< '(' function_arguments=any* ')' >
            >
        """
        )
        .modify(callback=handle_assertraises)
        .select(
            """
            function_call=power<
                self_attr="self" raises_attr=trailer< "." "assertRaisesRegex" >
                trailer< '(' function_arguments=any* ')' >
            >
        """
        )
        .modify(callback=handle_assertraises_regex)
        .select_method("assertRegex")
        .modify(callback=handle_assert_regex)
        .select_method("assertNotRegex")
        .modify(callback=handle_assert_regex)
        # Actually run all of the above.
        .execute(write=True, interactive=interactive, silent=silent)
    )


# TODO : Add this to fissix.fixer_util
def Assert(test, message=None, **kwargs):  # pylint: disable=invalid-name
    """Build an assertion statement"""
    if not isinstance(test, list):
        test = [test]
    test[0].prefix = " "
    if message is not None:
        if not isinstance(message, list):
            message = [message]
        message.insert(0, Comma())
        message[1].prefix = " "

    node = Node(
        syms.assert_stmt,
        [Leaf(TOKEN.NAME, "assert")] + test + (message or []),
        **kwargs,
    )
    return node


def conversion(func):
    """
    Decorator. Handle some boilerplate
    """

    @wraps(func)
    def wrapper(node, capture, filename):
        if capture.get("function_def"):
            # Not interested in `def assertEqual`, leave that alone.
            # We only care about *calls*
            return node

        arguments_nodes = capture["function_arguments"]
        if not arguments_nodes:
            return node

        # This is wrapped in a list for some reason?
        arguments_node = arguments_nodes[0]

        if arguments_node.type == syms.arglist:
            # multiple arguments
            actual_arguments = [n for n in arguments_node.children if n.type != TOKEN.COMMA]
        else:
            # one argument
            actual_arguments = [arguments_node]

        # Un-multi-line, where a and b are on separate lines
        actual_arguments = [a.clone() for a in actual_arguments]
        for arg in actual_arguments:
            arg.prefix = " "

        assertion = func(node, capture, actual_arguments)

        if assertion is not None:
            node.parent.set_child(0, assertion)
            node.remove()

    return wrapper


@conversion
def assertmethod_to_assert(node, capture, arguments):
    """
    self.assertEqual(foo, bar, msg)
    --> assert foo == bar, msg

    self.assertTrue(foo, msg)
    --> assert foo, msg

    self.assertIsNotNone(foo, msg)
    --> assert foo is not None, msg

    .. etc
    """
    function_name = capture["function_name"].value
    invert = function_name in INVERT_FUNCTIONS
    function_name = SYNONYMS.get(function_name, function_name)
    num_arguments = ARGUMENTS[function_name]

    if len(arguments) not in (num_arguments, num_arguments + 1):
        # Not sure what this is. Leave it alone.
        return None

    if len(arguments) == num_arguments:
        message = None
    else:
        message = arguments.pop()
        if message.type == syms.argument:
            # keyword argument (e.g. `msg=abc`)
            message = message.children[2].clone()

    if function_name == "assertIsInstance":
        arguments[0].prefix = ""
        assert_test_nodes = [Call(keyword("isinstance"), [arguments[0], Comma(), arguments[1]])]
        if invert:
            assert_test_nodes.insert(0, keyword("not"))
    elif function_name == "assertAlmostEqual":
        arguments[1].prefix = ""
        # TODO: insert the `import pytest` at the top of the file
        if invert:
            op_token = Leaf(TOKEN.NOTEQUAL, "!=", prefix=" ")
        else:
            op_token = Leaf(TOKEN.EQEQUAL, "==", prefix=" ")
        assert_test_nodes = [
            Node(
                syms.comparison,
                [
                    arguments[0],
                    op_token,
                    Node(
                        syms.power,
                        Attr(keyword("pytest"), keyword("approx", prefix=""))
                        + [
                            ArgList(
                                [
                                    arguments[1],
                                    Comma(),
                                    KeywordArg(keyword("abs"), Leaf(TOKEN.NUMBER, "1e-7")),
                                ]
                            )
                        ],
                    ),
                ],
            )
        ]
        # Adds a 'import pytest' if there wasn't one already
        touch_import(None, "pytest", node)

    else:
        op_tokens = OPERATORS[function_name]
        if not isinstance(op_tokens, list):
            op_tokens = [op_tokens]
        op_tokens = [o.clone() for o in op_tokens]

        if invert:
            if not op_tokens:
                op_tokens.append(keyword("not"))
            elif op_tokens[0].type == TOKEN.NAME and op_tokens[0].value == "is":
                op_tokens[0] = Node(syms.comp_op, [keyword("is"), keyword("not")], prefix=" ")
            elif op_tokens[0].type == TOKEN.NAME and op_tokens[0].value == "in":
                op_tokens[0] = Node(syms.comp_op, [keyword("not"), keyword("in")], prefix=" ")
            elif op_tokens[0].type == TOKEN.EQEQUAL:
                op_tokens[0] = Leaf(TOKEN.NOTEQUAL, "!=", prefix=" ")

        if num_arguments == 2:
            # a != b, etc.
            assert_test_nodes = [arguments[0]] + op_tokens + [arguments[1]]
        elif function_name == "assertTrue":
            assert_test_nodes = op_tokens + [arguments[0]]
            # not a
        elif function_name == "assertIsNone":
            # a is not None
            assert_test_nodes = [arguments[0]] + op_tokens
    return Assert(assert_test_nodes, message.clone() if message else None, prefix=node.prefix)


@conversion
def assertalmostequal_to_assert(node, capture, arguments):
    function_name = capture["function_name"].value
    invert = function_name in INVERT_FUNCTIONS
    function_name = SYNONYMS.get(function_name, function_name)

    nargs = len(arguments)
    if nargs < 2 or nargs > 5:
        return None

    def get_kwarg_value(index, name):
        idx = 0
        for arg in arguments:
            if arg.type == syms.argument:
                if arg.children[0].value == name:
                    return arg.children[2].clone()
            else:
                if idx == index:
                    return arg.clone()
                idx += 1
        return None

    first = get_kwarg_value(0, "first")
    second = get_kwarg_value(1, "second")

    if first is None or second is None:
        # Not sure what this is, leave it alone
        return

    places = get_kwarg_value(2, "places")
    msg = get_kwarg_value(3, "msg")
    delta = get_kwarg_value(4, "delta")

    if delta is not None:
        try:
            abs_delta = float(delta.value)
        except ValueError:
            # this should be a number, give up.
            return
    else:
        if places is None:
            places = 7
        else:
            try:
                places = int(places.value)
            except (ValueError, AttributeError):
                # this should be an int, give up.
                return
        abs_delta = "1e-%d" % places  # pylint: disable=consider-using-f-string

    arguments[1].prefix = ""
    if invert:
        op_token = Leaf(TOKEN.NOTEQUAL, "!=", prefix=" ")
    else:
        op_token = Leaf(TOKEN.EQEQUAL, "==", prefix=" ")
    assert_test_nodes = [
        Node(
            syms.comparison,
            [
                arguments[0],
                op_token,
                Node(
                    syms.power,
                    Attr(keyword("pytest"), keyword("approx", prefix=""))
                    + [
                        ArgList(
                            [
                                arguments[1],
                                Comma(),
                                KeywordArg(keyword("abs"), Leaf(TOKEN.NUMBER, abs_delta)),
                            ]
                        )
                    ],
                ),
            ],
        )
    ]
    # Adds a 'import pytest' if there wasn't one already
    touch_import(None, "pytest", node)

    return Assert(assert_test_nodes, msg.clone() if msg else None, prefix=node.prefix)


@conversion
def handle_assertraises(node, capture, arguments):
    """
    with self.assertRaises(x):

        --> with pytest.raises(x):

    self.assertRaises(ValueError, func, arg1)

        --> pytest.raises(ValueError, func, arg1)
    """
    capture["self_attr"].replace(keyword("pytest", prefix=capture["self_attr"].prefix))
    capture["raises_attr"].replace(Node(syms.trailer, [Dot(), keyword("raises", prefix="")]))
    # Let's remove the msg= keyword argument if found
    for child in node.children:
        if child.type != syms.trailer:
            continue
        for tchild in child.children:
            if tchild.type != syms.arglist:
                continue
            previous_argument = None
            for argument in list(tchild.children):
                if isinstance(argument, Leaf):
                    previous_argument = argument
                    continue
                if isinstance(argument, Node):
                    if argument.type != syms.argument:
                        previous_argument = argument
                        continue
                    for leaf in argument.leaves():
                        if leaf.value == "msg":
                            argument.remove()
                            if previous_argument.value == ",":
                                # previous_argument is a comma, remove it.
                                previous_argument.remove()

    # Adds a 'import pytest' if there wasn't one already
    touch_import(None, "pytest", node)


@conversion
def handle_assertraises_regex(node, capture, arguments):
    """
    with self.assertRaisesRegex(x, "regex match"):

        --> with pytest.raises(x, match="regex match"):

    self.assertRaises(ValueError, "regex match", func, arg1)

        --> pytest.raises(ValueError, func, arg1, match="regex match")
    """
    capture["self_attr"].replace(keyword("pytest", prefix=capture["self_attr"].prefix))
    capture["raises_attr"].replace(Node(syms.trailer, [Dot(), keyword("raises", prefix="")]))
    # Let's remove the msg= keyword argument if found and replace the second argument
    # with a match keyword argument
    regex_match = None
    for child in node.children:
        if child.type != syms.trailer:
            continue
        for tchild in child.children:
            if tchild.type != syms.arglist:
                continue
            previous_argument = None
            argnum = 0
            for argument in list(tchild.children):
                if isinstance(argument, Leaf):
                    if argument.value != ",":
                        argnum += 1
                    else:
                        previous_argument = argument
                        continue

                    if argnum != 2:
                        previous_argument = argument
                        continue

                    if argnum == 2:
                        regex_match = Node(
                            syms.argument,
                            [
                                Leaf(TOKEN.NAME, "match"),
                                Leaf(TOKEN.EQUAL, "="),
                                Leaf(TOKEN.STRING, argument.value),
                            ],
                            prefix=" ",
                        )
                        argument.remove()
                        if previous_argument and previous_argument.value == ",":
                            previous_argument.remove()
                        previous_argument = None
                        continue
                if isinstance(argument, Node):
                    if argument.type != syms.argument:
                        previous_argument = argument
                        continue
                    for leaf in argument.leaves():
                        if leaf.value == "msg":
                            argument.remove()
                            if previous_argument and previous_argument.value == ",":
                                # previous_argument is a comma, remove it.
                                previous_argument.remove()

    if regex_match:
        regex_match_added = False
        for child in node.children:
            if regex_match_added:
                break
            if child.type != syms.trailer:
                continue
            for tchild in child.children:
                if tchild.type != syms.arglist:
                    continue
                tchild.children.append(Leaf(TOKEN.COMMA, ","))
                tchild.children.append(regex_match)
                regex_match_added = True
                break

    # Adds a 'import pytest' if there wasn't one already
    touch_import(None, "pytest", node)


@conversion
def handle_assert_regex(node, capture, arguments):
    """
    self.assertRegex(text, pattern, msg)
    --> assert re.search(pattern, text), msg

    self.assertNotRegex(text, pattern, msg)
    --> assert not re.search(pattern, text), msg

    """
    function_name = capture["function_name"].value
    invert = function_name in INVERT_FUNCTIONS
    function_name = SYNONYMS.get(function_name, function_name)
    num_arguments = ARGUMENTS[function_name]

    if len(arguments) not in (num_arguments, num_arguments + 1):
        # Not sure what this is. Leave it alone.
        return None

    if len(arguments) == num_arguments:
        message = None
    else:
        message = arguments.pop()
        if message.type == syms.argument:
            # keyword argument (e.g. `msg=abc`)
            message = message.children[2].clone()

    arguments[0].prefix = " "
    arguments[1].prefix = ""

    # Adds a 'import re' if there wasn't one already
    touch_import(None, "re", node)

    op_tokens = []
    if invert:
        op_tokens.append(keyword("not"))

    assert_test_nodes = [
        Node(
            syms.power,
            op_tokens
            + Attr(keyword("re"), keyword("search", prefix=""))
            + [ArgList([arguments[1], Comma(), arguments[0]])],
        )
    ]
    return Assert(assert_test_nodes, message.clone() if message else None, prefix=node.prefix)
